import os
from sys import argv
from time import sleep
from datetime import datetime
from profiles.mygui import * 


# --- SECONDARY FUNCTIONS ---

EndOfSentence = ( '.', '!', '?', ';', ':', chr(8230) )


def Indicator(i,n):
    k = 10*i//n
    c = k*chr(9608)+(10-k)*chr(9617)
    return c


def Seconds():
    return (datetime.now()-datetime(1,1,1,0,0)).total_seconds()


def Duration(seconds):
    seconds = int(seconds)
    h = ((seconds // 3600))
    m = (seconds // 60) % 60
    s = seconds % 60
    if m < 10:
        m = str('0'+str(m))
    else:
        m = str(m)
    if s < 10:
        s = str('0'+str(s))
    else:
        s = str(s)
    return str(h)+':'+str(m)+':'+str(s)


# basic encodings: 'utf-8', 'utf-8-sig', 'utf-16', 'utf-16-le', 'utf-16-be', 'cp1252', 'cp1251'
def StrLoad(fname, *args):
    if len(args) == 0:
        f = open(fname,'r')
    else:
        f = open(fname, 'r', encoding=args[0])
    s = f.read()
    f.close()
    return s


def StrSave(s, fname, *args):
    if len(args) == 0:
        f = open(fname,'w')
    else:
        f = open(fname, 'w', encoding=args[0])
    f.write(s)
    f.close()
    return None


def FilesList(folder, ext_tuple):
    d = sorted(os.listdir(folder))
    a = []   
    for i in range(len(d)):
        if os.path.isfile(folder+'\\'+d[i]):
            _, file_extension = os.path.splitext(d[i])
            if file_extension in ext_tuple:
                a.append(d[i])
    return a


def RemoveFiles(folder, ext_tuple):
    filelist = os.listdir(folder)
    for f in filelist:
        ext = '.'+f.split('.')[-1]
        if ext in ext_tuple: os.remove(os.path.join(folder, f))    
    return None


# Waiting for a file to be available for a given number of seconds
def WaitUntilAccess(file_name, time_to_wait):
    time_counter = 0
    while not os.access(file_name,os.R_OK):
        sleep(0.100)
        time_counter += 0.100
        if time_counter > time_to_wait: break


# Adding SSML tags to text
def SSML(s):
    EndOfSentence = ( '.', '!', '?', ';', ':', chr(8230) )
    s = s.replace('&','')
    d = s.splitlines(False)
    c = ''
    for i in range(len(d)):
        for x in EndOfSentence:
            d[i] = str.replace(d[i],x+' ',x+'</s> <s>')
        d[i] = '<p><s>'+d[i]+'</s></p>'
        c += d[i]
    c = '<speak>'+c+'</speak>'
    return c, len(c)


def AddSSML(s):
    c, _ = SSML(s)
    return c


def XLen(s):
    _, L = SSML(s)
    return L  


# Cleaning a paragraph from extra characters
def CleanStr(s):
    # remove extra spaces inside
    while '  ' in s: s = s.replace('  ', ' ')  
    # three dots in a row are replaced by the symbol "ellipsis"
    s = s.replace('...',chr(8230))
    # simplify combinations of ! and ? 
    while '!!' in s: s = s.replace('!!', '!')  
    while '??' in s: s = s.replace('??', '?')
    s = s.strip()
    #if s[len(s)-1] not in EndOfSentence: s += '.' 
    return s


# Convert a text into a list of paragraphs
def TextToParagList(s):
    p = s.split('\n')
    a = [] 
    for i in range(len(p)): a.append(str.replace(p[i],'\r','\n'))
    # remove empty paragraphs; delete extra characters in non-empty paragraphs
    i = len(a)-1
    while i > -1:
        if len(a[i].strip()) == 0:
            del a[i]
        else:
            a[i] = CleanStr(a[i])
        i -= 1
    return a


# Splitting a paragraph in half without breaking sentences
def SplitHalf(s):
    LT = False
    k = len(s)//3
    z = 2*len(s)//3+1
    m = -1
    i = k
    # first try
    while i < z:
        if s[i] in EndOfSentence:
            m = i
            break
        i += 1
    # second try
    if m == -1:
        i = k
        while i < z:
            if s[i] in (','):
                m = i
                break
            i += 1
    # last try
    LT = True
    if m == -1:
        i = k
        while i < z:
            if s[i] in (' '):
                m = i
                break
            i += 1
    if LT:
        first = s[:m]
        last = s[m+1:]
    else:    
        first = s[:m+1]
        last = s[m+2:]
    d = ( first, last )
    return d


# Splitting large paragraphs in a list
def ParagsLimitation(a, maxlen):
    k = 0
    # more than enough for reasonably sized text
    while k < 100:
        need = False
        i = 0
        while i <= len(a)-1:
            if XLen(a[i]) > maxlen:
                need = True
                d = SplitHalf(a[i])
                a[i] = d[0]
                a.insert(i+1,d[1])
                i += 2
            i += 1
        if not need: break
        k += 1
    return a


# Combine adjacent small paragraphs into a single text
def ParagsUnion(a, maxlen):
    if len(a) == 0: return []
    b = []
    j = 0
    b.append(a[0])
    for i in range(1,len(a)):
        if XLen(b[j])+XLen(a[i])+1 < maxlen:
            b[j] = b[j]+'\n'+a[i]
        else:
            j += 1
            b.append(a[i])
    return b


def CreateJoinListsMP3(folder, n, startnum, *args):
    if len(args) > 0:
        name = args[0]+'_'
    else:
        name = ''
    g = FilesList(folder,('.mp3'))
    h = []
    for i in range(len(g)):
        if g[i][:4] == 'temp': h.append(g[i])
    c = ''
    k = startnum
    for i in range(len(h)): 
        c = c+'file '+"'"+folder+h[i]+"'"+'\n'
        if (i+1) % n == 0:
            StrSave(c,folder+name+str(k).zfill(5)+'.lst','utf-8')
            c = ''
            k +=1
        if i == len(h)-1:
            StrSave(c,folder+name+str(k).zfill(5)+'.lst','utf-8') 
    return None


# Apply dictionaries from dic_names list from dic_folder to text
# in i tab, and place the result in the j tab, making it active
def ApplyDics(i, j, dic_folder, dic_names):
    if len(dic_names) == 0: return None
    for k in range(len(dic_names)):
        if k == 0:
            StatusMessage(dic_names[k])
            WFilter(i,j,dic_folder+dic_names[k],True)
        else:
            StatusMessage(dic_names[k])
            WFilter(j,j,dic_folder+dic_names[k],True)
    StatusMessage('')
    WActive(j)
    return None


# Apply dictionaries from dic_names list from dic_folder to string s
def ApplyDicsToString(s, dic_folder, dic_names):
    if len(dic_names) == 0: return s
    WNew(-1,s)
    for k in range(len(dic_names)):
        StatusMessage(dic_names[k])
        WFilter(-1,-1,dic_folder+dic_names[k],True)
    s = WText(-1)
    StatusMessage('')
    return s  

# --- END OF SECONDARY FUNCTIONS ---
